"""
Module for tokenizing MaSzyna text format files.
"""

from io import StringIO
import re
from typing import Iterator


class TokenStream:
    """ Provides convenient access to tokens """
    def __init__(self, tokens_iterator: Iterator[str]):
        self._tokens_iterator = tokens_iterator

    def get_token(self) -> str:
        """ Get next token from stream.
        If there are no tokens, return None.
        """
        try:
            return next(self._tokens_iterator)
        except StopIteration:
            return None

    def get_tokens(self, amount: int) -> Iterator[str]:
        """ Get iterator with provided amount of tokens.
        If 'amount' is greater than total tokens amount in iterator,
        redundant tokens will be None.
        """
        return (self.get_token() for _ in range(amount))

    def get_tokens_before(self, endmark: str, max_length: int=None) -> Iterator[str]:
        """ Get iterator of all tokens before specified endmark.
        If endmark not exists in iterator, iterator will be yielding until end of stream.
        """
        # TODO: Add max_length parameter. It this value will be less than yielded tokens
        # amount, the Exception should be raised
        while True:
            token = self.get_token()

            if token == endmark or token is None:
                break

            yield token

    def __iter__(self):
        return self._tokens_iterator


class Tokenizer:
    def __init__(self):
        self._comments_map = {}
        self.register_comment('//', '\n')
        self.register_comment('/*', '*/')

        self._separators = set()
        self.register_separator(' ')
        self.register_separator(';')
        self.register_separator('\t')
        self.register_separator(',')

# -- Public API --

    def register_comment(self, startmark: str, endmark: str):
        """ Register comment delimiters """
        self._comments_map[startmark] = endmark
        self._update_comment_marks_regex()

    def unregister_comment(self, startmark: str):
        """ Unregister comment style with provided startmark """
        del self._comments_map[startmark]
        self._update_comment_marks_regex()

    def register_separator(self, separator: str):
        """ Register token separator """
        self._separators.add(separator)
        self._update_separators_regex()

    def unregister_separator(self, separator: str):
        """ Unregister provided token separator """
        try:
            self._separators.remove(separator)
            self._update_separators_regex()
        except KeyError:
            pass

    def tokenize(self, file_: str) -> TokenStream:
        """ Get TokenStream from file at given path """
        return TokenStream(self._tokenize(file_))

# -- Internals --

    def _tokenize(self, file_: StringIO) -> Iterator[str]:
        QUOTE = '\"'

        self._raw_tokens = self._tokenize_raw(file_)

        for token in self._raw_tokens:

            if token.startswith(QUOTE):
                # BUG: It not works for isolated quotes...
                token = self._expand_token_until(token, endmark=QUOTE)
                token = token.strip(QUOTE)

            comment_match = re.match(self._comment_start_regex, token)
            if comment_match:
                comment_startmark = comment_match.group(0)
                comment_endmark = self._comments_map[comment_startmark]
                self._expand_token_until(token, endmark=comment_endmark)
                continue

            token = token.strip('\n')

            if token != '':
                yield token.strip()

    def _tokenize_raw(self, file_: StringIO) -> Iterator[str]:
        for line in file_:
            line = self._prepare_line(line)

            for raw_token in line.split(' '):
                yield raw_token

    def _prepare_line(self, line) -> str:
        line = re.sub(self._separators_regex, ' ', line)
        # Isolate comment marks from regular tokens
        comment_marks = re.finditer(self._comment_marks_regex, line)
        if comment_marks:
            for mark_match in comment_marks:
                mark = mark_match.group(0)
                separated_mark = ' ' + mark + ' '
                line = line.replace(mark, separated_mark)
        # Remove redundant spaces
        line = re.sub(r' +', ' ', line).strip(' ')

        return line

    def _expand_token_until(self, token: str, endmark: str) -> str:
        try:
            while not token.endswith(endmark):
                next_token = next(self._raw_tokens)
                token_sep = '' if next_token == '\n' else ' '
                token += token_sep + next_token
        except StopIteration:
            pass
        finally:
            return token

    def _update_separators_regex(self):
        escaped_separators = (re.escape(sep) for sep in self._separators)
        # TODO: Some separators shouldn't be escaped, e.g \t.

        self._separators_regex = r'|'.join(escaped_separators)

    def _update_comment_marks_regex(self):
        self._update_comment_startmarks_regex()
        self._update_comment_endmarks_regex()
        self._comment_marks_regex = r'|'.join(
            (self._comment_start_regex, self._comment_end_regex)
        )

    def _update_comment_startmarks_regex(self):
        startmarks = self._comments_map.keys()
        escaped_marks = (re.escape(mark) for mark in startmarks)
        self._comment_start_regex = r'|'.join(escaped_marks)

    def _update_comment_endmarks_regex(self):
        endmarks = self._comments_map.values()
        escaped_marks = (re.escape(mark) for mark in endmarks)
        self._comment_end_regex = r'|'.join(escaped_marks).replace('\\\n', '\n')
